# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file implements common sparse / dense2sparse training algorithms."""

import dataclasses
import functools
import logging
from typing import Any, Callable, NamedTuple, Optional

import chex
import jax
import jax.numpy as jnp
import optax

from fjformer.jaxpruner import (
	mask_calculator,
	sparsity_distributions,
	sparsity_schedules,
	sparsity_types,
)

FilterFnType = sparsity_distributions.FilterFnType
TopkFnType = mask_calculator.TopKFnType
CustomSparsityMapType = sparsity_distributions.CustomSparsityMapType

SparsityDistributionFnType = Callable[
	[chex.ArrayTree],
	chex.ArrayTree,
]


class SparseState(NamedTuple):
	"""Holds sparsity related state and the original optimization state."""

	masks: chex.ArrayTree
	inner_state: Any = None
	target_sparsities: Optional[chex.ArrayTree] = None
	count: Any = None


@functools.partial(jax.jit, static_argnums=2)
def apply_mask(param, mask, is_packed=False):
	if mask is None:
		return param
	elif not is_packed:
		return param * mask
	else:
		unpacked_mask = jnp.unpackbits(mask)[: param.size]
		unpacked_mask = unpacked_mask.reshape(param.shape)
		return param * unpacked_mask


@dataclasses.dataclass
class BaseUpdater(object):
	"""Implements base class and the common update schedule."""

	scheduler: Any = sparsity_schedules.NoUpdateSchedule()
	# Whether to skip gradient updates when updating masks.
	skip_gradients: bool = False
	# TODO Create documentation for BaseUpdater.
	# Whether to zero out the gradients of pruned weights.
	is_sparse_gradients: bool = False
	# TODO Make sparsity default to be 0.
	sparsity_type: sparsity_types.SparsityType = sparsity_types.Unstructured()
	sparsity_distribution_fn: Optional[SparsityDistributionFnType] = functools.partial(
		sparsity_distributions.uniform, sparsity=0.0
	)
	rng_seed: Optional[jax.random.PRNGKey] = None
	use_packed_masks: bool = False
	topk_fn: Optional[TopkFnType] = None

	def __post_init__(self):
		"""Define variables."""
		# TODO Try enabling different sparsity types for different layers.
		if self.topk_fn is None:
			self.topk_fn = mask_calculator.get_topk_fn(self.sparsity_type)
		# We set default value None, so that we dont't need to initialize jax
		# to read the file. Without this we get error when running the tests.
		if self.rng_seed is None:
			self.rng_seed = jax.random.PRNGKey(8)

	def apply_masks(
		self,
		params,
		masks,
		is_packed=False,
	):
		"""Sparsifies parameters according to the masks.

		Args:
		  params: A tree of parameters.
		  masks: A tree of masks, the structure of the nested tree is identical to
		    that of params.
		  is_packed: A boolean, if True, masks are assumed to be packed into bits
		    and thus unpacked before sparsification.

		Returns:
		  A tree of masked parameters.
		"""
		if is_packed is None:
			is_packed = self.use_packed_masks
		return jax.tree_map(
			functools.partial(apply_mask, is_packed=is_packed), params, masks
		)

	def calculate_scores(
		self,
		params,
		sparse_state=None,
		grads=None,
		**kwargs,
	):
		"""Calculates sparsity scores of a given tree of parameters.

		Scores returned by this function is used to determining sparsity (e.g.,
		binary mask creation w/ top-k function). During training, this function is
		invoked by the `update_state` method which updates the sparse state during
		update iterations. This function is also invoked by the `instant_sparsify`
		method.

		Args:
		  params: A tree of parameters.
		  sparse_state: An optional sparse state.
		  grads: An optional tree of gradients.
		  **kwargs: Additional arguments that child classes can consume.

		Returns:
		  A tree of the sparsity scores of parameters.
		Raises:
		  NotImplemeterError.
		"""
		raise NotImplementedError

	def get_initial_masks(self, params, target_sparsities):
		"""Generate initial mask. This is only used when .wrap_optax() is called."""

		def mask_fn(p, target):
			if target is None:
				return None
			return jnp.ones(p.shape, dtype=mask_calculator.MASK_DTYPE)

		masks = jax.tree_map(mask_fn, params, target_sparsities)
		return masks

	def create_masks(self, scores, sparsities):
		def topk_ifnot_none(score, sparsity):
			return None if sparsity is None else self.topk_fn(score, sparsity)

		return jax.tree_map(topk_ifnot_none, scores, sparsities)

	def init_state(self, params):
		"""Creates the sparsity state."""
		if self.sparsity_distribution_fn is None:
			target_sparsities = None
		else:
			target_sparsities = self.sparsity_distribution_fn(params)
		logging.info("target_sparsities: %s", target_sparsities)
		masks = self.get_initial_masks(params, target_sparsities)
		if self.use_packed_masks:
			masks = jax.tree_map(jnp.packbits, masks)
		return SparseState(
			masks=masks,
			target_sparsities=target_sparsities,
			count=jnp.zeros([], jnp.int32),
		)

	def update_state(
		self,
		sparse_state,
		params,
		grads,
		**kwargs,
	):
		"""Update masks in sparse_state according to the current parameters and gradients.

		Args:
		  sparse_state: A wrapped optimization state. Original optimization state is
		    stored under `.inner`.
		  params: A tree of parameters.
		  grads: A tree of gradients.
		  **kwargs: Additional arguments to pass to `calculate_scores`.

		Returns:
		  A sparse state with the updated masks tree.
		"""
		sparsities = self.scheduler.get_sparsity_at_step(
			sparse_state.target_sparsities, sparse_state.count
		)
		scores = self.calculate_scores(
			params, sparse_state=sparse_state, grads=grads, **kwargs
		)
		new_masks = self.create_masks(scores, sparsities)
		if self.use_packed_masks:
			new_masks = jax.tree_map(jnp.packbits, new_masks)
		return sparse_state._replace(masks=new_masks)

	def wrap_optax(self, inner):
		"""Wraps an existing optax transformation and adds sparsity related updates.

		The gradient transformation provided (`inner`) is called as it is at
		initialization and during the update steps. In addition to this, a sparse
		state is created using the `self.init_state` function, which includes
		variables like masks needed by the algorithms. The sparse state is updated
		according to the given schedule using the `self.update_state` function.

		The returned transformation is a GradientTransformationExtraArgs: its
		update function accepts arbitrary additional args which it passes to
		`self.update_state`.

		Args:
		  inner: An optax gradient transformation.

		Returns:
		  An updated optax gradient transformation.
		"""

		def init_fn(params):
			sparse_state = self.init_state(params)
			sparse_state = sparse_state._replace(inner_state=inner.init(params))
			return sparse_state

		def update_fn(updates, state, params, **kwargs):
			is_update_step = self.scheduler.is_mask_update_iter(state.count)
			no_update_op = lambda state, *_: state  # noqa
			new_state = jax.lax.cond(
				is_update_step,
				functools.partial(self.update_state, **kwargs),
				no_update_op,
				state,
				params,
				updates,
			)

			if self.is_sparse_gradients:
				new_updates = self.apply_masks(updates, new_state.masks)
			else:
				new_updates = updates

			should_skip_inner = jnp.logical_and(is_update_step, self.skip_gradients)

			def no_inner_update(updates, inner_state, params):
				# Set gradients to zero and don't update the step.
				del params
				zero_updates = jax.tree_map(jnp.zeros_like, updates)
				return zero_updates, inner_state

			new_updates, new_inner_state = jax.lax.cond(
				should_skip_inner,
				no_inner_update,
				inner.update,
				new_updates,
				new_state.inner_state,
				params,
			)
			new_state = new_state._replace(
				count=optax.safe_int32_increment(new_state.count),
				inner_state=new_inner_state,
			)
			return new_updates, new_state

		new_tx = optax.GradientTransformationExtraArgs(init_fn, update_fn)
		new_tx.__pruner_tx__ = True  # modified
		return new_tx

	def instant_sparsify(
		self,
		params,
		grads=None,
		**kwargs,
	):
		"""Instantly sparsifies parameters with the sparsity distribution provided at initialization.

		Args:
		  params: A tree of parameters.
		  grads: An optional tree of gradients.
		  **kwargs: Additional arguments to pass to `calculate_scores`.

		Returns:
		  Either pruned parameters or a tuple of pruned parameters and masks.
		"""
		if self.sparsity_distribution_fn is None:
			raise ValueError(
				"sparsity_distribution_fn needs to be provided for sparsification."
			)
		sparsities = self.sparsity_distribution_fn(params)
		scores = self.calculate_scores(params, grads=grads, **kwargs)
		masks = self.create_masks(scores, sparsities)
		masked_params = self.apply_masks(params, masks, is_packed=False)
		if self.use_packed_masks:
			masks = jax.tree_map(jnp.packbits, masks)
		return masked_params, masks

	def pre_forward_update(self, params, opt_state):
		"""Used to transform paramaters before forward pass."""
		del opt_state
		return params

	def post_gradient_update(self, params, sparse_state):
		"""Used to transform parameters after gradient updates.

		The default behavior of this function is to sparsifiy parameters with masks
		after gradient updates.

		Args:
		  params: A tree of parameters.
		  sparse_state: A sparse state containing masks.

		Returns:
		  A tree of pruned parameters.
		"""
		return self.apply_masks(params, sparse_state.masks, is_packed=self.use_packed_masks)


@dataclasses.dataclass
class NoPruning(BaseUpdater):
	"""Implements no-operation updater, which keeps the parameters dense."""

	sparsity_distribution_fn: Optional[SparsityDistributionFnType] = None

	def get_initial_masks(self, params, target_sparsities):
		del target_sparsities
		return jax.tree_map(lambda p: None, params)

	def update_state(self, sparse_state, params, grads):
		"""Identity operation, returns the state unmodified."""
		return sparse_state

	def instant_sparsify(self, params, grads=None):
		"""Returns parameters as it is, together w/ None masks."""
		return params, self.get_initial_masks(params, None)



# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for utility functions in mask_calculator.py."""

import jax
import jax.numpy as jnp
from absl.testing import absltest, parameterized

from fjformer.jaxpruner import mask_calculator, sparsity_types

SCORE_1 = [
	[
		[0.69, 0.37, 0.96, 0.81],
		[0.79, 0.55, 0.88, 0.53],
		[0.66, 0.88, 0.4, 0.83],
		[0.13, 0.8, 0.91, 0.96],
	],
	[
		[0.34, 0.17, 0.07, 0.74],
		[0.13, 0.13, 0.83, 0.76],
		[0.13, 0.92, 0.25, 0.75],
		[0.37, 0.42, 0.65, 0.46],
	],
]
MASK_UNS_1 = [
	[[0, 0, 1, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 1, 1]],
	[[0, 0, 0, 0], [0, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 0]],
]
MASK_N2_M4_1 = [
	[[0, 0, 1, 1], [1, 0, 1, 0], [0, 1, 0, 1], [0, 0, 1, 1]],
	[[1, 0, 0, 1], [0, 0, 1, 1], [0, 1, 0, 1], [0, 0, 1, 1]],
]
MASK_N2_M4_AXIS1_1 = [
	[[1, 0, 1, 0], [1, 0, 0, 0], [0, 1, 0, 1], [0, 1, 1, 1]],
	[[1, 0, 0, 0], [0, 0, 1, 1], [0, 1, 0, 1], [1, 1, 1, 0]],
]
MASK_CHANNEL_AXIS1_1 = [
	[
		[0, 0, 0, 0],
		[0, 0, 0, 0],
		[1, 1, 1, 1],
		[1, 1, 1, 1],
	],
	[
		[0, 0, 0, 0],
		[0, 0, 0, 0],
		[1, 1, 1, 1],
		[1, 1, 1, 1],
	],
]
MASK_CHANNEL_AXIS2_1 = [
	[
		[0, 0, 1, 1],
		[0, 0, 1, 1],
		[0, 0, 1, 1],
		[0, 0, 1, 1],
	],
	[
		[0, 0, 1, 1],
		[0, 0, 1, 1],
		[0, 0, 1, 1],
		[0, 0, 1, 1],
	],
]

SCORE_2 = [1, 1, 1, 1, 1, 1, 1, 1]
MASK_UNS_2 = [1, 1, 1, 1, 0, 0, 0, 0]
MASK_N2_M4_2 = [0, 0, 1, 1, 0, 0, 1, 1]

SCORE_3 = [1, 2, 3, 4, 5, 6, 7, 8]

SCORE_4 = [1, 0.5, 0.3, 0.2, 1, 1.5, 1.8, 2.0, 1, 1]
MASK_UNS_4 = [1, 0, 0, 0, 1, 1, 1, 1, 0, 0]

SCORE_5 = [
	[1, 0.5, 0.3, 0.2, 1, 1.5, 1.8, 2.0, 1, 1],
	[1, 0.5, 0.3, 0.2, 1, 1.5, 1.8, 2.0, 1, 1],
]
MASK_UNS_5 = [[1, 0, 0, 0, 1, 1, 1, 1, 1, 1], [0, 0, 0, 0, 0, 1, 1, 1, 0, 0]]
MASK_CHANNEL_AXIS1_5 = [
	[1, 0, 0, 0, 1, 1, 1, 1, 0, 0],
	[1, 0, 0, 0, 1, 1, 1, 1, 0, 0],
]


class MaskCalculatorTest(parameterized.TestCase, absltest.TestCase):
	@parameterized.parameters(
		(SCORE_1, 0.8, MASK_UNS_1),
		(SCORE_2, 0.5, MASK_UNS_2),
		(SCORE_4, 0.5, MASK_UNS_4),
		(SCORE_5, 0.5, MASK_UNS_5),
	)
	def testScoreBasedTopKWithUnstructuredInput(self, score, sparsity, expected_mask):
		topk_fn = mask_calculator.get_topk_fn(sparsity_types.Unstructured)
		topk_fn_in_jit = jax.jit(topk_fn)
		self.assertTrue(
			jnp.array_equal(topk_fn(jnp.array(score), sparsity), jnp.array(expected_mask))
		)
		self.assertTrue(
			jnp.array_equal(
				topk_fn_in_jit(jnp.array(score), sparsity), jnp.array(expected_mask)
			)
		)

	@parameterized.product(jit_compile=[False, True], pool_type=["AVG", "MAX"])
	def testScoreBasedTopKWithBlockInput2dScores(self, jit_compile, pool_type):
		sparsity_type = sparsity_types.Block(
			block_shape=(2, 3),
			use_avg_pooling=(pool_type == "AVG"),
		)
		topk_fn = mask_calculator.get_topk_fn(sparsity_type)
		if jit_compile:
			topk_fn = jax.jit(topk_fn)

		scores = [
			[0.69, 0.37, 0.96, 0.81, 0.79, 0.55, 0.88, 0.53],
			[0.66, 0.88, 0.4, 0.83, 0.13, 0.8, 0.91, 0.96],
			[0.34, 0.17, 0.07, 0.74, 0.13, 0.13, 0.83, 0.76],
			[0.13, 0.92, 0.25, 0.75, 0.37, 0.42, 0.65, 0.46],
		]
		mask = topk_fn(scores, sparsity=0.2)

		expected_mask_avg = [
			[1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
			[1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
			[0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0],
			[0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0],
		]
		expected_mask_max = [
			[1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
			[1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
			[1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 1.0, 1.0],
			[1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 1.0, 1.0],
		]

		if pool_type == "AVG":
			self.assertTrue(jnp.array_equal(mask, jnp.array(expected_mask_avg)))
		else:
			self.assertTrue(jnp.array_equal(mask, jnp.array(expected_mask_max)))

	@parameterized.product(jit_compile=[False, True], pool_type=["AVG", "MAX"])
	def testScoreBasedTopKWithBlockInput1dScores(self, jit_compile, pool_type):
		sparsity_type = sparsity_types.Block(
			block_shape=(1, 3),
			use_avg_pooling=(pool_type == "AVG"),
		)
		topk_fn = mask_calculator.get_topk_fn(sparsity_type)
		if jit_compile:
			topk_fn = jax.jit(topk_fn)

		scores = [0.69, 0.37, 0.96, 0.81, 0.79, 0.55, 0.88, 0.53]
		mask = topk_fn(scores, sparsity=0.2)

		expected_mask_avg = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0]
		expected_mask_max = [1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 1.0, 1.0]

		if pool_type == "AVG":
			self.assertTrue(jnp.array_equal(mask, jnp.array(expected_mask_avg)))
		else:
			self.assertTrue(jnp.array_equal(mask, jnp.array(expected_mask_max)))

	@parameterized.product(jit_compile=[False, True], pool_type=["AVG", "MAX"])
	def testScoreBasedTopKWithBlockInput3dScores(self, jit_compile, pool_type):
		sparsity_type = sparsity_types.Block(
			block_shape=(2, 2),
			use_avg_pooling=(pool_type == "AVG"),
		)
		topk_fn = mask_calculator.get_topk_fn(sparsity_type)
		if jit_compile:
			topk_fn = jax.jit(topk_fn)

		scores = [
			[[0.69, 0.37], [0.96, 0.81], [0.79, 0.55], [0.88, 0.53]],
			[[0.66, 0.88], [0.4, 0.83], [0.13, 0.8], [0.91, 0.96]],
			[[0.34, 0.17], [0.07, 0.74], [0.13, 0.13], [0.83, 0.76]],
			[[0.13, 0.92], [0.25, 0.75], [0.37, 0.42], [0.65, 0.46]],
		]
		mask = topk_fn(scores, sparsity=0.2)

		expected_mask_avg = [
			[[1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [1.0, 1.0]],
			[[1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [1.0, 1.0]],
			[[0.0, 0.0], [0.0, 0.0], [1.0, 1.0], [1.0, 1.0]],
			[[0.0, 0.0], [0.0, 0.0], [1.0, 1.0], [1.0, 1.0]],
		]
		expected_mask_max = [
			[[1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [1.0, 1.0]],
			[[1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [1.0, 1.0]],
			[[1.0, 1.0], [1.0, 1.0], [0.0, 0.0], [0.0, 0.0]],
			[[1.0, 1.0], [1.0, 1.0], [0.0, 0.0], [0.0, 0.0]],
		]

		if pool_type == "AVG":
			self.assertTrue(jnp.array_equal(mask, jnp.array(expected_mask_avg)))
		else:
			self.assertTrue(jnp.array_equal(mask, jnp.array(expected_mask_max)))

	@parameterized.parameters(
		(SCORE_1, sparsity_types.NByM(n=2, m=4), MASK_N2_M4_1),
		(SCORE_2, sparsity_types.NByM(n=2, m=4), MASK_N2_M4_2),
		(SCORE_3, sparsity_types.NByM(n=4, m=2), None),
		(SCORE_1, sparsity_types.NByM(n=2, m=4, axis=1), MASK_N2_M4_AXIS1_1),
		(SCORE_3, sparsity_types.NByM(n=4, m=2, axis=0), None),
	)
	def testScoreBasedTopKWithNbyMInput(self, score, sparsity_type, expected_mask):
		topk_fn = mask_calculator.get_topk_fn(sparsity_type)
		topk_fn_in_jit = jax.jit(topk_fn)

		if sparsity_type.n > sparsity_type.m:
			with self.assertRaises(ValueError):
				topk_fn(jnp.array(score), 0.5)
				topk_fn_in_jit(jnp.array(score), 0.5)
			return

		self.assertTrue(
			jnp.array_equal(topk_fn(jnp.array(score), 0.5), jnp.array(expected_mask))
		)
		self.assertTrue(
			jnp.array_equal(
				topk_fn_in_jit(jnp.array(score), 0.5),
				jnp.array(expected_mask),
			)
		)

	@parameterized.parameters(
		dict(shape=[18, 4000, 4000], axis=-1),
		dict(shape=[18, 4000, 4000], axis=-2),
	)
	def testTopKWithNbyMLimitMemoryUsage(self, shape, axis):
		# `score`` takes about 2.304 GB (4 * 36 * 4000 * 4000 / 10**6) of memory.
		# Since the size of HBM is 8 GB for each core of Jellyfish. This test only
		# pass if the peak memory does not exceed 8GB of memory.
		score = jax.random.normal(jax.random.PRNGKey(0), shape)
		topk_fn = mask_calculator.get_topk_fn(sparsity_types.NByM(n=2, m=4, axis=axis))
		topk_fn_in_jit = jax.jit(topk_fn)
		topk_fn_in_jit(score, 0.5)

	@parameterized.parameters(
		dict(shape=[2, 4, 4], axis=0),
		dict(shape=[2, 5, 3], axis=1),
	)
	def testTopKWithNbyMTargetAxisNotDivisibleByM(self, shape, axis):
		score = jax.random.normal(jax.random.PRNGKey(0), shape)
		topk_fn = mask_calculator.get_topk_fn(sparsity_types.NByM(n=2, m=4, axis=axis))
		with self.assertRaises(ValueError):
			topk_fn(score, 0.5)

	@parameterized.parameters(
		dict(shape=[2, 4, 4], axis=1),
		dict(shape=[2, 4, 8], axis=2),
	)
	def testTopKWithNbyMGenerateNonPrunedMaskForSparsityZero(self, shape, axis):
		score = jax.random.normal(jax.random.PRNGKey(0), shape)
		topk_fn = mask_calculator.get_topk_fn(sparsity_types.NByM(n=2, m=4, axis=axis))
		non_pruned_mask = topk_fn(score, 0)
		pruned_mask = topk_fn(score, 0.5)

		self.assertEqual((non_pruned_mask == 0).mean(), 0)
		self.assertEqual((pruned_mask == 0).mean(), 0.5)

	@parameterized.parameters(
		(SCORE_1, 0.5, sparsity_types.Channel(axis=1), MASK_CHANNEL_AXIS1_1),
		(SCORE_1, 0.5, sparsity_types.Channel(axis=-2), MASK_CHANNEL_AXIS1_1),
		(SCORE_1, 0.5, sparsity_types.Channel(), MASK_CHANNEL_AXIS2_1),
		(SCORE_1, 0.5, sparsity_types.Channel(axis=2), MASK_CHANNEL_AXIS2_1),
		(SCORE_1, 0.5, sparsity_types.Channel(axis=-1), MASK_CHANNEL_AXIS2_1),
		(SCORE_5, 0.5, sparsity_types.Channel(axis=1), MASK_CHANNEL_AXIS1_5),
	)
	def testScoreBasedTopKWithChannelInput(
		self, score, sparsity, sparsity_type, expected_mask
	):
		topk_fn = mask_calculator.get_topk_fn(sparsity_type)
		topk_fn_in_jit = jax.jit(topk_fn)

		self.assertTrue(
			jnp.array_equal(topk_fn(jnp.array(score), sparsity), jnp.array(expected_mask))
		)
		self.assertTrue(
			jnp.array_equal(
				topk_fn_in_jit(jnp.array(score), sparsity),
				jnp.array(expected_mask),
			)
		)


if __name__ == "__main__":
	absltest.main()



"""Utility functions for JAX."""

import importlib.util
import logging
from typing import Tuple, Union

import flax
import jax
from jax import random as jrandom


def is_torch_available() -> bool:
	"""Checks if PyTorch is installed.

	Returns:
	    True if the torch module is installed, False otherwise.
	"""
	return importlib.util.find_spec("torch") is not None


def count_num_params(params: flax.core.frozen_dict.FrozenDict) -> int:
	"""Counts the number of parameters in a model.

	Args:
	    params: A Flax FrozenDict containing the model parameters.

	Returns:
	    The total number of parameters in the model.
	"""
	return sum(i.size for i in jax.tree_util.tree_flatten(flax.core.unfreeze(params))[0])


def count_params(params: flax.core.frozen_dict.FrozenDict) -> None:
	"""Prints the number of parameters in a Flax model.

	Args:
	    params: A Flax FrozenDict containing the model parameters.
	"""
	print(
		f"\033[1;31mModel Contains : {count_num_params(params) / 1e9:.2f} Billion Parameters"
	)


class JaxRNG:
	"""A wrapper around JAX's PRNGKey that simplifies key splitting."""

	def __init__(self, rng: jrandom.PRNGKey):
		"""Initializes the JaxRNG with a PRNGKey.

		Args:
		    rng: A JAX PRNGKey.
		"""
		self.rng = rng

	@classmethod
	def from_seed(cls, seed: int) -> "JaxRNG":
		"""Creates a JaxRNG instance from a seed.

		Args:
		    seed: The seed to use for the random number generator.

		Returns:
		    A JaxRNG instance.
		"""
		return cls(jrandom.PRNGKey(seed))

	def __call__(
		self, keys: Union[int, Tuple[str, ...]] = None
	) -> Union[jrandom.PRNGKey, Tuple[jrandom.PRNGKey, ...], dict]:
		"""Splits the internal PRNGKey and returns new keys.

		Args:
		    keys:  If None, returns a single split key and updates the internal RNG.
		           If an int, splits the key into `keys + 1` parts, updates the internal RNG,
		           and returns the last `keys` parts as a tuple.
		           If a tuple of strings, splits the key into `len(keys) + 1` parts,
		           updates the internal RNG, and returns a dictionary mapping the strings
		           to their corresponding key parts.

		Returns:
		    The split PRNGKey(s) based on the `keys` argument.
		"""
		if keys is None:
			self.rng, split_rng = jrandom.split(self.rng)
			return split_rng
		elif isinstance(keys, int):
			split_rngs = jrandom.split(self.rng, num=keys + 1)
			self.rng = split_rngs[0]
			return tuple(split_rngs[1:])
		else:
			split_rngs = jrandom.split(self.rng, num=len(keys) + 1)
			self.rng = split_rngs[0]
			return {key: val for key, val in zip(keys, split_rngs[1:])}  # noqa:B905


# Global JaxRNG instance
jax_utils_rng = None


def init_rng(seed: int) -> None:
	"""Initializes the global JaxRNG with a seed.

	Args:
	    seed: The seed to use for the random number generator.
	"""
	global jax_utils_rng
	jax_utils_rng = JaxRNG.from_seed(seed)


def next_rng(
	*args, **kwargs
) -> Union[jrandom.PRNGKey, Tuple[jrandom.PRNGKey, ...], dict]:
	"""Provides access to the global JaxRNG and splits the key based on arguments.

	This function wraps the global `jax_utils_rng` instance and calls its `__call__` method,
	passing through any arguments provided. This provides a convenient way to access and
	split the global random number generator key.

	Args:
	    *args: Positional arguments passed to the `jax_utils_rng` instance's `__call__` method.
	    **kwargs: Keyword arguments passed to the `jax_utils_rng` instance's `__call__` method.

	Returns:
	    The split PRNGKey(s) from the global `jax_utils_rng` instance.
	"""
	global jax_utils_rng
	return jax_utils_rng(*args, **kwargs)


class GenerateRNG:
	"""An infinite generator of JAX PRNGKeys, useful for iterating over seeds."""

	def __init__(self, seed: int = 0):
		"""Initializes the generator with a starting seed.

		Args:
		    seed: The seed to use for the initial PRNGKey.
		"""
		self.seed = seed
		self._rng = jrandom.PRNGKey(seed)

	def __next__(self) -> jrandom.PRNGKey:
		"""Generates and returns the next PRNGKey in the sequence.

		Returns:
		    The next PRNGKey derived from the internal state.
		"""
		self._rng, key = jrandom.split(self._rng)
		return key

	@property
	def rng(self) -> jrandom.PRNGKey:
		"""Provides access to the next PRNGKey without advancing the generator.

		Returns:
		    The next PRNGKey in the sequence.
		"""
		return next(self)


def get_logger(name, level: int = logging.INFO) -> logging.Logger:
	"""
	Function to create and configure a logger.
	Args:
	    name: str: The name of the logger.
	    level: int: The logging level. Defaults to logging.INFO.
	Returns:
	    logging.Logger: The configured logger instance.
	"""
	logger = logging.getLogger(name)
	logger.propagate = False

	# Set the logging level
	logger.setLevel(level)

	# Create a console handler
	console_handler = logging.StreamHandler()
	console_handler.setLevel(level)

	formatter = logging.Formatter("%(asctime)s %(levelname)-8s [%(name)s] %(message)s")
	console_handler.setFormatter(formatter)
	logger.addHandler(console_handler)
	return logger


def set_loggers_level(level: int = logging.WARNING):
	"""Function to set the logging level of all loggers to the specified level.

	Args:
	    level: int: The logging level to set. Defaults to
	        logging.WARNING.
	"""
	logging.root.setLevel(level)
	for handler in logging.root.handlers:
		handler.setLevel(level)



# coding=utf-8
# Copyright 2024 Jaxpruner Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for api functions in api.py."""

import inspect

import jaxpruner
import ml_collections
from absl.testing import absltest

from fjformer.jaxpruner import api, sparsity_schedules, sparsity_types


class ScenicConfigTest(absltest.TestCase):
	def testCreateUpdater(self):
		sparsity_config = ml_collections.ConfigDict()
		sparsity_config.algorithm = "magnitude"
		sparsity_config.update_freq = 10
		sparsity_config.update_start_step = 200
		sparsity_config.update_end_step = 1000
		sparsity_config.sparsity = 0.8
		sparsity_config.dist_type = "erk"
		updater = api.create_updater_from_config(sparsity_config)

		self.assertIsInstance(updater, jaxpruner.MagnitudePruning)

		self.assertIsInstance(updater.sparsity_type, sparsity_types.Unstructured)

		self.assertIsInstance(updater.scheduler, sparsity_schedules.PolynomialSchedule)
		self.assertEqual(updater.scheduler.update_freq, sparsity_config.update_freq)
		self.assertEqual(
			updater.scheduler.update_start_step, sparsity_config.update_start_step
		)
		self.assertEqual(updater.scheduler.update_end_step, sparsity_config.update_end_step)

		func_args = inspect.signature(updater.sparsity_distribution_fn).parameters
		self.assertLen(func_args, 4)
		self.assertIn("param_tree", func_args)
		self.assertEqual(func_args["sparsity"].default, sparsity_config.sparsity)

	def testCreateUpdaterNoPruning(self):
		sparsity_config = ml_collections.ConfigDict()
		sparsity_config.algorithm = "no_prune"
		updater = api.create_updater_from_config(sparsity_config)

		self.assertIsInstance(updater, jaxpruner.NoPruning)

	def testCreateUpdaterErrorWithWrongDistFn(self):
		sparsity_config = ml_collections.ConfigDict()
		sparsity_config.algorithm = "magnitude"
		sparsity_config.dist_type = "not_a_distribution_function"

		with self.assertRaisesRegex(ValueError, "dist_type:"):
			api.create_updater_from_config(sparsity_config)

	def testCreateUpdaterErrorWithWrongType(self):
		sparsity_config = ml_collections.ConfigDict()
		sparsity_config.algorithm = "not_a_correct_algorithm"
		sparsity_config.dist_type = "uniform"
		sparsity_config.sparsity = 0.9

		with self.assertRaisesRegex(ValueError, "jaxpruner.all_algorithm_names()"):
			api.create_updater_from_config(sparsity_config)

	def testCreateUpdaterErrorWithWrongSparsityType(self):
		sparsity_config = ml_collections.ConfigDict()
		sparsity_config.algorithm = "magnitude"
		sparsity_config.dist_type = "uniform"
		sparsity_config.sparsity = 0.4
		sparsity_config.sparsity_type = "wrong_sparsity_type"

		with self.assertRaisesRegex(ValueError, "Sparsity type"):
			api.create_updater_from_config(sparsity_config)

	def testCreateUpdaterNoUpdateScheduleWithNoStartStep(self):
		sparsity_config = ml_collections.ConfigDict()
		sparsity_config.algorithm = "magnitude"
		sparsity_config.dist_type = "uniform"
		sparsity_config.sparsity = 0.9
		updater = api.create_updater_from_config(sparsity_config)

		self.assertIsInstance(updater.scheduler, sparsity_schedules.NoUpdateSchedule)


if __name__ == "__main__":
	absltest.main()



"""
This module implements the ImplicitArray class and related functionality for
representing abstract arrays without instantiation.
It provides mechanisms for lazy evaluation and efficient handling of array operations
in JAX.

Key components:
- ImplicitArray: Abstract base class for symbolic array representations
- primitive_handler: Decorator for registering custom primitive handlers
- implicit_compact: Decorator for functions to accept ImplicitArray arguments
"""

import warnings
from abc import ABC, abstractmethod
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass, field, fields, is_dataclass
from functools import partial, wraps
from itertools import chain, count
from typing import Any, Callable, ClassVar, Optional, Tuple

import jax
import jax.extend.linear_util as lu
import jax.interpreters.partial_eval as pe
from jax import core, tree_util
from jax._src.typing import DTypeLike, Shape
from jax.api_util import flatten_fun, flatten_fun_nokwargs
from jax.tree_util import register_pytree_with_keys_class
from plum import Dispatcher, Function

from fjformer.core.errors import UninitializedAval

_dispatch = Dispatcher()
_primitive_ids = count()
warnings.filterwarnings(
	"ignore", message="Could not resolve the type hint of `~B`", module="plum.type"
)
warnings.filterwarnings(
	"ignore", message="Could not resolve the type hint of `~A`", module="plum.type"
)


class ArrayValue(ABC):  # noqa
	"""
	Abstract ArrayValue to register jax.Arrays
	"""


ArrayValue.register(jax.Array)


def get_lax_primitive_by_name(name: str) -> jax.core.Primitive:
	"""Get a JAX LAX primitive by its name."""
	return getattr(jax.lax, f"{name}_p")


def get_primitive_handler(primitive):
	"""Get or create a handler for a given primitive."""
	if isinstance(primitive, str):
		primitive = get_lax_primitive_by_name(primitive)
	handler = _dispatch.functions.get(primitive)
	if handler is None:

		def _not_impl_handler(primitive: jax.core.Primitive, *args, **kwargs):
			return NotImplemented

		_not_impl_handler.__doc__ = f"Default handler for {primitive.name}"
		handler = Function(_not_impl_handler)
		handler.register(_not_impl_handler, precedence=-1e9)
		handler.__name__ = f"{primitive.name}_{next(_primitive_ids)}"
		_dispatch.functions[primitive] = handler
	return handler


def primitive_handler(primitives, precedence=0):
	"""Decorator to register a handler for one or more primitives."""
	if isinstance(primitives, (str, jax.core.Primitive)):
		primitives = [primitives]

	def decorator(fn):
		for primitive in primitives:
			handler = get_primitive_handler(primitive)
			handler.register(fn, precedence=precedence)

	return decorator


COMMUTATIVE_OPS = frozenset(
	[
		"add",
		"bitwise_and",
		"bitwise_or",
		"bitwise_xor",
		"eq",
		"max",
		"min",
		"mul",
		"ne",
	]
)

ELEMENTWISE_UNOPS = frozenset(
	[
		"abs",
		"acos",
		"acosh",
		"asin",
		"asinh",
		"atan",
		"atanh",
		"bessel_i0e",
		"bessel_i1e",
		"cbrt",
		"ceil",
		"clz",
		"conj",
		"convert_element_type",
		"copy",
		"cos",
		"cosh",
		"digamma",
		"erf_inv",
		"erf",
		"erfc",
		"exp",
		"expm1",
		"floor",
		"imag",
		"integer_pow",
		"is_finite",
		"lgamma",
		"log1p",
		"log",
		"logistic",
		"neg",
		"not",
		"population_count",
		"real",
		"reduce_precision",
		"round",
		"rsqrt",
		"sign",
		"sin",
		"sinh",
		"sqrt",
		"tan",
		"tanh",
	]
)

ELEMENTWISE_BINOPS = frozenset(
	[
		"add",
		"and",
		"atan2",
		"complex",
		"div",
		"eq",
		"ge",
		"gt",
		"igamma_grad_a",
		"igamma",
		"igammac",
		"le",
		"lt",
		"max",
		"min",
		"mul",
		"ne",
		"nextafter",
		"or",
		"pow",
		"random_gamma_grad",
		"rem",
		"shift_left",
		"shift_right_arithmetic",
		"shift_right_logical",
		"sub",
		"xor",
	]
)

REDUCTION_OPS = frozenset(
	[
		"argmax",
		"argmin",
		"reduce_and",
		"reduce_max",
		"reduce_min",
		"reduce_or",
		"reduce_prod",
		"reduce_sum",
		"reduce_xor",
	]
)

CUMULATIVE_REDUCTION_OPS = frozenset(
	[
		"cumlogsumexp",
		"cummax",
		"cummin",
		"cumprod",
		"cumsum",
	]
)


class _EmptyNodeCls:
	_instance = None

	def __new__(cls):
		if cls._instance is None:
			cls._instance = super().__new__(cls)
		return cls._instance


EmptyNode = _EmptyNodeCls()

tree_util.register_pytree_node(
	_EmptyNodeCls, lambda node: ((), None), lambda _, __: EmptyNode
)


def combine_leaf_predicate(base_fn, is_leaf):
	@wraps(base_fn)
	def new_fn(*args, new_is_leaf=None):
		if new_is_leaf is None:
			combined_is_leaf = is_leaf
		else:

			def combined_is_leaf(arg):
				return is_leaf(arg) or new_is_leaf(arg)

		return base_fn(*args, is_leaf=combined_is_leaf)

	return new_fn


def leaf_predicate(x):
	return isinstance(x, (ImplicitArray, _EmptyNodeCls))


tree_map_with_implicit = combine_leaf_predicate(jax.tree_util.tree_map, leaf_predicate)
tree_map_with_path_with_implicit = combine_leaf_predicate(
	tree_util.tree_map_with_path, leaf_predicate
)
tree_flatten_with_implicit = combine_leaf_predicate(
	tree_util.tree_flatten, leaf_predicate
)
tree_flatten_with_path_with_implicit = combine_leaf_predicate(
	tree_util.tree_flatten_with_path, leaf_predicate
)
tree_leaves_with_implicit = combine_leaf_predicate(
	tree_util.tree_leaves, leaf_predicate
)
tree_structure_with_implicit = combine_leaf_predicate(
	tree_util.tree_structure, leaf_predicate
)


def flatten_one_implicit_layer(tree):
	def is_leaf_below_node(node, x):
		return isinstance(x, ImplicitArray) and x is not node

	def replace_subtree_implicits(node):
		return tree_util.tree_map(
			lambda _: 1, node, is_leaf=partial(is_leaf_below_node, node)
		)

	prototype = tree_map_with_implicit(replace_subtree_implicits, tree)
	struct = tree_util.tree_structure(prototype)

	leaves = tree_leaves_with_implicit(tree)
	leaves = list(
		chain.from_iterable(
			(
				tree_util.tree_leaves(leaf, is_leaf=partial(is_leaf_below_node, leaf))
				if isinstance(leaf, ImplicitArray)
				else [leaf]
			)
			for leaf in leaves
		)
	)
	return leaves, struct


def implicit_depth(tree):
	leaves = tree_leaves_with_implicit(tree)
	depth = 0
	while True:
		next_leaves = []
		any_implicit = False
		for leaf in leaves:
			if not isinstance(leaf, ImplicitArray):
				continue
			any_implicit = True
			next_leaves.extend(flatten_one_implicit_layer(leaf)[0])

		if not any_implicit:
			return depth

		depth += 1
		leaves = next_leaves


def _map_leaves_with_implicit_path(f, leaves, is_leaf, path_prefix=()):
	mapped_leaves = []
	for idx, leaf in enumerate(leaves):
		path = path_prefix + (idx,)
		if not isinstance(leaf, ImplicitArray) or is_leaf(path, leaf):
			mapped_leaves.append(f(leaf))
			continue

		subtree, substruct = flatten_one_implicit_layer(leaf)
		mapped_subtree = _map_leaves_with_implicit_path(
			f, subtree, is_leaf=is_leaf, path_prefix=path
		)
		mapped_leaves.append(tree_util.tree_unflatten(substruct, mapped_subtree))
	return mapped_leaves


def _get_pruning_transform(tree, materialization_paths):
	if not materialization_paths:
		return lambda x: x

	def is_leaf(path, leaf):
		return path in materialization_paths

	def materialize_subtrees(tree):
		leaves, struct = tree_flatten_with_implicit(tree)

		mapped_leaves = _map_leaves_with_implicit_path(
			partial(materialize_nested, full=True), leaves, is_leaf
		)
		return tree_util.tree_unflatten(struct, mapped_leaves)

	return materialize_subtrees


def get_common_prefix_transforms(trees):
	"""
	Given an iterable of pytrees which have the same structure after all
	ImplicitArray instances are materialized, return a list of callables
	which will transform each tree into the largest common structure
	obtainable via materialization of ImplicitArrays.
	"""
	if len(trees) <= 1:
		return [lambda x: x for _ in trees]

	all_leaves, structures = zip(  # noqa: B905
		*(tree_flatten_with_implicit(tree) for tree in trees)
	)
	post_materialization_avals = [core.get_aval(leaf) for leaf in all_leaves[0]]
	for i, (leaves, structure) in enumerate(
		zip(all_leaves[1:], structures[1:]),  # noqa: B905
		1,
	):
		if structure != structures[0]:
			raise ValueError("Trees do not have the same structure after materialization")

		for leaf, expected_aval in zip(  # noqa: B905
			leaves, post_materialization_avals
		):
			aval = core.get_aval(leaf)
			if not (aval.shape == expected_aval.shape and aval.dtype == expected_aval.dtype):
				raise ValueError(
					f"Trees do not have the same avals after materialization. Tree 0: {expected_aval}, Tree {i}: {aval}"
				)

	# Stack will contain tuples of (path, nodes)
	# path = a sequence of integers specifying which child
	# was taken at each _flatten_one_implicit_layer call
	# or the first flatten_with_implicit call
	# nodes = one node from each tree
	stack = []

	all_leaves = []
	for tree in trees:
		all_leaves.append(tree_leaves_with_implicit(tree))

	for i, nodes in enumerate(zip(*all_leaves)):  # noqa: B905
		stack.append(((i,), nodes))

	materialization_paths = set()
	while stack:
		path_prefix, nodes = stack.pop()
		if not any(isinstance(node, ImplicitArray) for node in nodes):
			continue

		all_leaves, all_structures = zip(  # noqa: B905
			*(flatten_one_implicit_layer(node) for node in nodes)
		)
		node_structures = set(all_structures)
		if len(node_structures) > 1:
			materialization_paths.add(path_prefix)
			continue

		aval_diff = False
		for leaves in zip(*all_leaves):  # noqa: B905
			first_aval = core.get_aval(leaves[0])
			shape = first_aval.shape
			dtype = first_aval.dtype
			for leaf in leaves[1:]:
				aval = core.get_aval(leaf)
				if not (aval.shape == shape and aval.dtype == dtype):
					materialization_paths.add(path_prefix)
					aval_diff = True
			if aval_diff:
				break

		if aval_diff:
			continue

		for i, leaf_group in enumerate(zip(*all_leaves)):  # noqa: B905
			stack.append((path_prefix + (i,), leaf_group))

	return [_get_pruning_transform(tree, materialization_paths) for tree in trees]


def materialize_nested(implicit_arr, full=False):
	"""
	Materialize an ImplicitArray instance, handling the case where implicit_arr.materialize()
	involves further ImplicitArray instances.
	Arguments:
	    implicit_arr: An ImplicitArray instance
	    full: If True, repeatedly materialize until the result is a concrete array
	Returns:
	    The materialized array
	"""
	while isinstance(implicit_arr, ImplicitArray):
		wrapped = lu.wrap_init(type(implicit_arr).materialize)
		flat, in_tree = flatten_one_implicit_layer((implicit_arr,))
		flat_fn, out_tree = flatten_fun_nokwargs(wrapped, in_tree)
		out_flat = implicit_compact(flat_fn.call_wrapped)(*flat)
		implicit_arr = jax.tree_util.tree_unflatten(out_tree(), out_flat)

		if not full:
			break

	return implicit_arr


def _with_implicit_flat(fun: lu.WrappedFun) -> lu.WrappedFun:
	"""Wrap a function to handle implicit arrays."""
	f = _implicit_inner(fun)
	return _implicit_outer(f)


@lu.transformation
def _implicit_outer(*in_vals):
	"""Outer transformation for implicit array handling."""
	with core.new_main(ImplicitArrayTrace) as main:
		outs = yield (main, *in_vals), {}
		del main
	yield outs


@lu.transformation
def _implicit_inner(main, *in_vals):
	"""Inner transformation for implicit array handling."""
	trace = main.with_cur_sublevel()
	in_tracers = [
		ImplicitArrayTracer(trace, val) if isinstance(val, ImplicitArray) else val
		for val in in_vals
	]
	outs = yield in_tracers, {}
	out_vals = [trace.full_raise(t).value for t in outs]
	yield out_vals


def implicit_compact(f: Callable) -> Callable:
	"""
	A decorator that enables compact handling of ImplicitArray subclasses within a function.
	This allows for seamless integration of custom array types in JAX operations.

	This decorator can be used in combination with jax.jit for optimized execution.

	Args:
	    f: The function to be decorated.

	Returns:
	    A wrapped function that can handle both regular arrays and ImplicitArray instances.

	Example:
	    >>> @jax.jit
	    >>> @implicit_compact
	    >>> def f(a, b):
	    ...   return jnp.dot(a, b)

	    >>> result = f(regular_array, regular_array)
	    >>> implicit_result = f(implicit_array, implicit_or_normal_array)
	"""

	@wraps(f)
	def implicit_f(*args, **kwargs):
		flat_args, in_tree = tree_flatten_with_implicit((args, kwargs))
		f_flat, out_tree = flatten_fun(lu.wrap_init(f), in_tree)
		f_wrapped = _with_implicit_flat(f_flat)
		outs_flat = f_wrapped.call_wrapped(*flat_args)
		return out_tree().unflatten(outs_flat)

	return implicit_f


def default_handler(primitive: Any, *args: Any, **params: Any) -> Any:
	"""Default handler for primitives."""
	subfuns, bind_params = primitive.get_bind_params(params)
	return primitive.bind(*subfuns, *args, **bind_params)


def aux_field(metadata: Optional[dict] = None, **kwargs: Any) -> Any:
	"""Create an auxiliary field for ImplicitArray subclasses."""
	metadata = dict(metadata) if metadata else {}
	metadata["implicit_array_aux"] = True
	return field(metadata=metadata, **kwargs)


class _AvalDescriptor:
	"""Descriptor for lazy initialization of shape and dtype."""

	def __set_name__(self, owner: Any, name: str) -> None:
		self._name = f"_{name}"

	def __get__(self, obj: Any, owner: Any = None) -> Any:
		if obj is None:
			return None
		result = getattr(obj, self._name, None)
		if result is None:
			raise UninitializedAval(kind=self._name[1:])
		return result

	def __set__(self, obj: Any, value: Any) -> None:
		setattr(obj, self._name, value)


# Context variable for aval discovery
_aval_discovery = ContextVar("aval_discovery", default=False)


@contextmanager
def _aval_discovery_context():
	"""Context manager for aval discovery."""
	token = _aval_discovery.set(True)
	try:
		yield
	finally:
		_aval_discovery.reset(token)


@dataclass
class _ImplicitArrayBase(ArrayValue, ABC):
	"""Base class for ImplicitArray with common attributes."""

	commute_ops: ClassVar[bool] = True
	warn_on_materialize: ClassVar[bool] = True
	default_shape: ClassVar[Optional[Tuple[int, ...]]] = None
	default_dtype: ClassVar[Optional[Any]] = None

	shape: Optional[Tuple[int, ...]] = aux_field(kw_only=True, default=None)
	dtype: Any = aux_field(kw_only=True, default=None)


@dataclass
class ImplicitArray(_ImplicitArrayBase):
	"""
	Abstract class for representing an abstract array without instantiation.

	Subclasses must implement the materialize method, which defines the relationship
	between the implicit array and the value it represents. Subclasses are valid
	arguments to functions decorated with core.implicit_compact.

	The represented shape and dtype may be defined in various ways:
	1. Explicitly passing shape/dtype keyword arguments at initialization
	2. Overriding the default_shape/default_dtype class variables
	3. Overriding the compute_shape/compute_dtype methods
	4. Overriding __post_init__ and manually setting shape/dtype
	5. Inferred by running jax.eval_shape() on the subclass's materialize method
	"""

	shape = _AvalDescriptor()
	dtype = _AvalDescriptor()

	def __post_init__(self):
		"""Initialize shape and dtype after instance creation."""
		try:
			aval = _get_materialization_aval(self)
		except UninitializedAval:
			aval = None

		shape = None
		try:
			shape = self.shape
		except UninitializedAval:
			shape = self.shape = self.compute_shape()

		if aval is not None:
			if shape is None:
				self.shape = aval.shape
			elif shape != aval.shape:
				warnings.warn(
					f"ImplicitArray shape {shape} does not match materialization shape {aval.shape}",
					stacklevel=1,
				)
		elif shape is None:
			raise UninitializedAval("shape")

		dtype = None
		try:
			dtype = self.dtype
		except UninitializedAval:
			dtype = self.dtype = self.compute_dtype()

		if dtype is None and aval is None:
			aval = _get_materialization_aval(self)

		if aval is not None:
			if dtype is None:
				self.dtype = aval.dtype
			elif dtype != aval.dtype:
				warnings.warn(
					f"ImplicitArray dtype {dtype} does not match materialization dtype {aval.dtype}",
					stacklevel=1,
				)
		elif dtype is None:
			raise UninitializedAval("dtype")

	def compute_shape(self) -> Optional[Shape]:
		"""Compute the shape of the implicit array."""
		return self.default_shape

	def compute_dtype(self) -> Optional[DTypeLike]:
		"""Compute the dtype of the implicit array."""
		return self.default_dtype

	@property
	def aval(self) -> core.ShapedArray:
		"""Get the abstract value of the implicit array."""
		return core.ShapedArray(self.shape, self.dtype)

	@classmethod
	def default_handler(cls, primitive, *args, params=None):
		"""Default handler for primitives."""
		if params is None:
			params = {}
		return materialize_handler(primitive, *args, params=params)

	@abstractmethod
	def materialize(self):
		"""Materialize the implicit array into a concrete array."""
		pass

	def tree_flatten_with_keys(self):
		"""Flatten the implicit array for PyTree operations."""
		children = []
		aux_data = []
		for name, is_aux in _get_names_and_aux(self):
			try:
				value = getattr(self, name)
			except UninitializedAval:
				if not _aval_discovery.get():
					raise
				value = None
			if is_aux:
				aux_data.append(value)
			else:
				children.append((name, value))

		return children, aux_data

	@classmethod
	def tree_unflatten(cls, aux_data, children):
		"""Unflatten the implicit array from PyTree operations."""
		child_it = iter(children)
		aux_it = iter(aux_data)
		obj = cls.__new__(cls)
		for name, is_aux in _get_names_and_aux(cls):
			value = next(aux_it if is_aux else child_it)
			setattr(obj, name, value)

		return obj

	def handle_primitive(self, primitive, *args, params):
		"""Handle a primitive operation on the implicit array."""
		handler = lu.wrap_init(partial(get_primitive_handler(primitive), primitive))
		use_params = params

		if len(args) == 2 and self.commute_ops:
			args, use_params = _maybe_swap_args(primitive.name, args, use_params)

		flat_args, in_tree = flatten_one_implicit_layer((args, params))
		flat_handler, out_tree = flatten_fun(handler, in_tree)

		result = implicit_compact(flat_handler.call_wrapped)(*flat_args)
		return jax.tree_util.tree_unflatten(out_tree(), result)

	def __init_subclass__(cls, commute_ops=True, warn_on_materialize=True, **kwargs):
		"""Initialize a subclass of ImplicitArray."""
		super().__init_subclass__(**kwargs)
		cls.commute_ops = commute_ops
		cls.warn_on_materialize = warn_on_materialize

		if not is_dataclass(cls):
			raise TypeError(f"{cls.__name__} must be a dataclass")
		core.pytype_aval_mappings[cls] = lambda x: x.aval
		register_pytree_with_keys_class(cls)
		return cls


def _get_names_and_aux(obj):
	for val in fields(obj):
		yield val.name, bool(val.metadata.get("implicit_array_aux"))


def _materialize_all(it):
	return [
		materialize_nested(val) if isinstance(val, ImplicitArray) else val for val in it
	]


def _maybe_swap_args(op_name, args, params):
	if isinstance(args[0], ImplicitArray):
		return args, params
	if op_name in COMMUTATIVE_OPS:
		return args[::-1], params

	return args, params


class ImplicitArrayTracer(core.Tracer):
	def __init__(self, trace, value):
		super().__init__(trace)
		self.value = value

	@property
	def aval(self):
		if isinstance(self.value, ImplicitArray):
			return self.value.aval
		return core.get_aval(self.value)

	def full_lower(self):
		if isinstance(self.value, ImplicitArray):
			return self

		return core.full_lower(self.value)


class ImplicitArrayTrace(core.Trace):
	pure = lift = lambda self, val: ImplicitArrayTracer(self, val)

	def process_primitive(self, primitive, tracers, params):
		outs = NotImplemented
		vals = [t.value for t in tracers]
		implicit_idx = next(i for i, v in enumerate(vals) if isinstance(v, ImplicitArray))

		# First try to handle the primitive using custom handlers
		outs = vals[implicit_idx].handle_primitive(primitive, *vals, params=params)

		if outs is NotImplemented:
			# For higher order primitives most users won't implement custom
			# logic, so there shouldn't be a warning
			if primitive.name in _default_handlers:
				outs = _default_handlers[primitive.name](primitive, *vals, params=params)
			else:
				implicit_cls = vals[implicit_idx].__class__
				if implicit_cls.warn_on_materialize:
					warnings.warn(
						f"Primitive {primitive.name} was not handled by class {implicit_cls.__name__}, so implicit args will be materialized.",
						stacklevel=1,
					)

		if outs is NotImplemented:
			outs = vals[implicit_idx].default_handler(primitive, *vals, params=params)

		if primitive.multiple_results:
			return [ImplicitArrayTracer(self, out) for out in outs]
		return ImplicitArrayTracer(self, outs)


def wrap_jaxpr(jaxpr, vals_with_implicits, return_closed=True):
	if isinstance(jaxpr, jax.core.ClosedJaxpr):
		literals = jaxpr.literals
		jaxpr = jaxpr.jaxpr
	else:
		literals = []

	wrapped_fn = lu.wrap_init(implicit_compact(partial(core.eval_jaxpr, jaxpr)))
	flat_args, in_tree = jax.tree_util.tree_flatten((literals, *vals_with_implicits))
	flat_fn, out_tree = flatten_fun_nokwargs(wrapped_fn, in_tree)

	new_jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(
		flat_fn, [core.get_aval(v) for v in flat_args]
	)

	ret = (
		(jax.core.ClosedJaxpr(new_jaxpr, consts),) if return_closed else (new_jaxpr, consts)
	)
	return *ret, flat_args, out_tree()


def _transform_jaxpr_output(jaxpr, jaxpr_args, orig_out_struct, out_transform):
	def eval_fn(literals, *args):
		output = implicit_compact(partial(core.eval_jaxpr, jaxpr.jaxpr))(literals, *args)
		unflattened_output = orig_out_struct.unflatten(output)
		return out_transform(unflattened_output)

	wrapped = lu.wrap_init(eval_fn)

	flat_args, in_tree = jax.tree_util.tree_flatten((jaxpr.literals, *jaxpr_args))
	flat_fn, out_tree = flatten_fun_nokwargs(wrapped, in_tree)
	new_jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(
		flat_fn, [core.get_aval(v) for v in flat_args]
	)

	return jax.core.ClosedJaxpr(new_jaxpr, consts), out_tree()


def _match_branches(branches, arg_vals):
	out_avals = []
	new_jaxprs = []
	flat_inputs = None
	branch_out_struct = None
	for branch in branches:
		new_jaxpr, flat_inputs, branch_out_struct = wrap_jaxpr(branch, arg_vals)
		new_jaxprs.append((new_jaxpr, branch_out_struct))
		out_avals.append(
			branch_out_struct.unflatten(
				jax.eval_shape(
					partial(core.eval_jaxpr, new_jaxpr.jaxpr),
					new_jaxpr.literals,
					*flat_inputs,
				)
			)
		)

	out_transforms = get_common_prefix_transforms(out_avals)
	new_branches = []
	out_struct = None
	for (new_jaxpr, orig_out_struct), transform in zip(new_jaxprs, out_transforms):  # noqa
		new_jaxpr, out_struct = _transform_jaxpr_output(
			new_jaxpr, flat_inputs, orig_out_struct, transform
		)
		new_branches.append(new_jaxpr)

	return tuple(new_branches), out_struct, flat_inputs


def _handle_cond(primitive, *vals, params):
	cond_val, *arg_vals = vals
	subfuns, bind_params = primitive.get_bind_params(params)

	new_branches, out_struct, flat_inputs = _match_branches(params["branches"], arg_vals)
	bind_params["branches"] = new_branches
	bind_params["linear"] = _broadcast_tuple(bind_params["linear"], arg_vals)

	outs = primitive.bind(*subfuns, cond_val, *flat_inputs, **bind_params)
	return jax.tree_util.tree_unflatten(out_struct, outs)


def _handle_remat2(primitive, *vals, params):
	subfuns, bind_params = primitive.get_bind_params(params)
	new_jaxpr, consts, flat_inputs, out_tree = wrap_jaxpr(
		bind_params["jaxpr"], vals, return_closed=False
	)
	new_jaxpr = pe.convert_constvars_jaxpr(new_jaxpr)
	bind_params["jaxpr"] = new_jaxpr
	outs = primitive.bind(*subfuns, *consts, *flat_inputs, **bind_params)
	return jax.tree_util.tree_unflatten(out_tree, outs)


def _handle_pjit(primitive, *vals, params):
	new_jaxpr, flat_inputs, out_tree = wrap_jaxpr(params["jaxpr"], vals)
	donated_invars = _broadcast_tuple(params["donated_invars"], vals)
	in_shardings = _broadcast_tuple(params["in_shardings"], vals)
	out_shardings = _broadcast_tuple(params["out_shardings"], out_tree)

	subfuns, bind_params = primitive.get_bind_params(params)
	bind_params["jaxpr"] = new_jaxpr
	bind_params["donated_invars"] = donated_invars
	bind_params["in_shardings"] = in_shardings
	bind_params["out_shardings"] = out_shardings
	outs = primitive.bind(*subfuns, *flat_inputs, **bind_params)
	return jax.tree_util.tree_unflatten(out_tree, outs)


def _handle_scan(primitive, *vals, params):
	n_consts = params["num_consts"]
	n_carry = params["num_carry"]

	consts = vals[:n_consts]
	real_n_consts = len(jax.tree_util.tree_leaves(consts))

	carries = vals[n_consts : n_consts + n_carry]
	xs = vals[n_consts + n_carry :]

	if any(isinstance(c, ImplicitArray) for c in carries):
		warnings.warn("Not Supported Yet.", stacklevel=1)
		carries = _materialize_all(carries)

	sliced_xs = jax.tree_util.tree_map(partial(jax.eval_shape, lambda x: x[0]), xs)

	for x in sliced_xs:
		if isinstance(x, ImplicitArray):
			assert len(x._shape) > 0, "Attempted to scan over a scalar."
			x._shape = x._shape[1:]

	jaxpr = params["jaxpr"]
	new_jaxpr, _, out_tree = wrap_jaxpr(
		jaxpr=jaxpr,
		vals_with_implicits=(*consts, *carries, *sliced_xs),
		return_closed=True,
	)

	flat_inputs = jax.tree_util.tree_leaves((jaxpr.literals, *consts, *carries, *xs))

	subfuns, bind_params = primitive.get_bind_params(params)
	bind_params["jaxpr"] = new_jaxpr
	bind_params["num_consts"] = real_n_consts
	bind_params["num_carry"] = len(carries)
	bind_params["linear"] = _broadcast_tuple(params["linear"], vals)

	outs = primitive.bind(*subfuns, *flat_inputs, **bind_params)
	return jax.tree_util.tree_unflatten(out_tree, outs)


_default_handlers = {
	"cond": _handle_cond,
	"remat2": _handle_remat2,
	"pjit": _handle_pjit,
	"scan": _handle_scan,
}


def materialize_handler(primitive, *vals, params):
	vals = _materialize_all(vals)
	subfuns, bind_params = primitive.get_bind_params(params)
	result = implicit_compact(primitive.bind)(*subfuns, *vals, **bind_params)
	return result


def _broadcast_tuple(t, trees):
	if isinstance(trees, jax.tree_util.PyTreeDef):
		trees = jax.tree_util.tree_unflatten(trees, range(trees.num_leaves))
	assert len(t) == len(trees)
	return tuple(
		chain.from_iterable(
			(tuple_val for _ in jax.tree_util.tree_leaves(tree))
			for tuple_val, tree in zip(t, trees)  # noqa
		)
	)


def _get_materialization_aval(imp_arr):
	with _aval_discovery_context(), _filter_materialization_warnings():
		result = jax.eval_shape(partial(materialize_nested, full=True), imp_arr)
	return result


@contextmanager
def _filter_materialization_warnings():
	with warnings.catch_warnings():
		warnings.filterwarnings("ignore", message="Primitive.*was not handled")
		yield


# Most Core part's of this package is Developed from orginal implimentation of QAX from davisyoshida



# TODO : Implement Custom Backward Prp
from dataclasses import dataclass
from functools import partial
from typing import Any, Literal, Optional, Sequence, Union

import jax
import triton
from jax import Array, lax
from jax import numpy as jnp
from jax.core import Primitive
from triton import language as tl

import fjformer.core as core
from fjformer.jax_triton import strides_from_shape, triton_call, cdiv


def get_gpu_plat():
	target = jax.devices()[0].device_kind.lower()
	if "nvidia" in target:
		return "cuda"
	elif "amd" in target:
		return "rocm"
	return None


match get_gpu_plat():
	case "cuda":

		@triton.jit
		def trround(x):
			return tl.extra.cuda.libdevice.rint(x)
	case _:

		@triton.jit
		def trround(x):
			return tl.floor(x + 0.5)


# @triton.autotune(
# 	[
# 		triton.Config({}, num_warps=16, num_stages=2),
# 		triton.Config({}, num_warps=8, num_stages=2),
# 		triton.Config({}, num_warps=4, num_stages=2),
# 		triton.Config({}, num_warps=2, num_stages=2),
# 		triton.Config({}, num_warps=1, num_stages=2),
# 	],
# 	key=["K"],
# )
@triton.jit
def quantize_row_q8_triton(
	A,
	M,
	K,
	stride_am,
	stride_ak,
	stride_qm,
	stride_qk,
	stride_sm,
	Q,
	S,
	MCache: tl.constexpr,
	BLOCK_SIZE_M: tl.constexpr,
	BLOCK_SIZE_K: tl.constexpr,
):
	pid_m = tl.program_id(axis=0)
	A_Block_ptr = tl.make_block_ptr(
		base=A,
		shape=(M, K),
		block_shape=(BLOCK_SIZE_M, BLOCK_SIZE_K),
		offsets=(pid_m * BLOCK_SIZE_M, 0),
		strides=(stride_am, stride_ak),
		order=(0, 1),
	)
	Q_Block_ptr = tl.make_block_ptr(
		base=Q,
		shape=(M, K),
		block_shape=(BLOCK_SIZE_M, BLOCK_SIZE_K),
		offsets=(pid_m * BLOCK_SIZE_M, 0),
		strides=(stride_qm, stride_qk),
		order=(0, 1),
	)
	S_Block_ptr = tl.make_block_ptr(
		base=S,
		shape=(M,),
		block_shape=(BLOCK_SIZE_M,),
		offsets=(pid_m * BLOCK_SIZE_M,),
		strides=(stride_sm,),
		order=(0,),
	)
	a = tl.load(A_Block_ptr)
	scales = tl.max(tl.abs(a), axis=1) / 127.0
	doted = a * tl.where(scales > 0, 1 / scales, 0)[:, None]
	quant = trround(doted).to(tl.int8)
	tl.store(Q_Block_ptr, quant)
	tl.store(S_Block_ptr, scales)


# @triton.autotune(
# 	[
# 		triton.Config({}, num_warps=16, num_stages=2),
# 		triton.Config({}, num_warps=8, num_stages=2),
# 		triton.Config({}, num_warps=4, num_stages=2),
# 		triton.Config({}, num_warps=2, num_stages=2),
# 		triton.Config({}, num_warps=1, num_stages=2),
# 	],
# 	key=["K"],
# )
@triton.jit
def dequantize_row_q8_triton(
	Q,
	S,
	M,
	K,
	stride_am,
	stride_ak,
	stride_qm,
	stride_qk,
	stride_sm,
	A,
	MCache: tl.constexpr,
	BLOCK_SIZE_M: tl.constexpr,
	BLOCK_SIZE_K: tl.constexpr,
):
	pid_m = tl.program_id(axis=0)
	pid_k = tl.program_id(axis=1)
	A_Block_ptr = tl.make_block_ptr(
		base=A,
		shape=(M, K),
		block_shape=(BLOCK_SIZE_M, BLOCK_SIZE_K),
		offsets=(pid_m * BLOCK_SIZE_M, pid_k * BLOCK_SIZE_K),
		strides=(stride_am, stride_ak),
		order=(0, 1),
	)
	Q_Block_ptr = tl.make_block_ptr(
		base=Q,
		shape=(M, K),
		block_shape=(BLOCK_SIZE_M, BLOCK_SIZE_K),
		offsets=(pid_m * BLOCK_SIZE_M, pid_k * BLOCK_SIZE_K),
		strides=(stride_qm, stride_qk),
		order=(0, 1),
	)
	S_Block_ptr = tl.make_block_ptr(
		base=S,
		shape=(M,),
		block_shape=(BLOCK_SIZE_M,),
		offsets=(pid_m * BLOCK_SIZE_M,),
		strides=(stride_sm,),
		order=(0,),
	)
	quants = tl.load(Q_Block_ptr)
	scale = tl.load(S_Block_ptr)
	out = tl.zeros([BLOCK_SIZE_M, BLOCK_SIZE_K], tl.float32)
	out += quants * scale[:, None]
	tl.store(A_Block_ptr, out)


def quantize_row_q8_triton_call(array):
	assert array.ndim == 2
	M, K = array.shape
	BLOCK_SIZE_M = 128
	BLOCK_SIZE_K = triton.next_power_of_2(K)

	quants_shape = jax.ShapeDtypeStruct((M, K), jnp.int8)
	scales_shape = jax.ShapeDtypeStruct((M,), jnp.float32)

	stride_am, stride_ak = strides_from_shape(array.shape)
	stride_qm, stride_qk = strides_from_shape(quants_shape.shape)
	stride_sm = strides_from_shape(scales_shape.shape)

	quants, scales = triton_call(
		array,
		M,
		K,
		stride_am,
		stride_ak,
		stride_qm,
		stride_qk,
		stride_sm,
		kernel=quantize_row_q8_triton,
		MCache=cdiv(M, 2048),
		BLOCK_SIZE_M=BLOCK_SIZE_M,
		BLOCK_SIZE_K=BLOCK_SIZE_K,
		grid=(cdiv(M, BLOCK_SIZE_M), 1, 1),
		out_shape=[quants_shape, scales_shape],
		num_warps=8,
		num_stages=2,
	)

	return quants, scales.reshape(M, 1).astype(jnp.float16)


def dequantize_row_q8_triton_call(quants, scales):
	assert quants.ndim == 2
	M, K = quants.shape
	scales = scales.reshape(-1)
	BLOCK_SIZE_M = 128
	BLOCK_SIZE_K = triton.next_power_of_2(K)

	array_shape = jax.ShapeDtypeStruct((M, K), jnp.float32)

	stride_am, stride_ak = strides_from_shape(array_shape.shape)
	stride_qm, stride_qk = strides_from_shape(quants.shape)
	stride_sm = strides_from_shape(scales.shape)

	(array,) = triton_call(
		quants,
		scales,
		M,
		K,
		stride_am,
		stride_ak,
		stride_qm,
		stride_qk,
		stride_sm,
		kernel=dequantize_row_q8_triton,
		MCache=cdiv(M, 2048),
		BLOCK_SIZE_M=BLOCK_SIZE_M,
		BLOCK_SIZE_K=BLOCK_SIZE_K,
		grid=(cdiv(M, BLOCK_SIZE_M), 1, 1),
		out_shape=[array_shape],
		num_warps=8,
		num_stages=2,
	)

	return array


@jax.jit
def _mu_quantize_row_q8_0(x):
	"""
	Quantize a row of float32 values to 8-bit integers with blockwise scaling.
	Args:
	    x: input array
	Returns:
	    tuple of (scales, quantized_values)
	    - scales: float16 array of shape (nb,)
	    - quantized_values: int8 array of shape (k,)
	"""
	amax = jnp.max(jnp.abs(x), axis=-1, keepdims=True)
	d = amax / 127.0
	ids = jnp.where(d > 0, 1.0 / d, 0.0)
	x_scaled = x * ids
	quantized = jnp.round(x_scaled)
	quantized = quantized.astype(jnp.int8)
	return quantized, d.astype(jnp.float16)


@jax.jit
def _mu_dequantize_row_q8_0(quants, scales):
	"""
	Dequantize 8-bit integers back to float32 values using blockwise scaling.

	Args:
	    quants: int8 array of shape (k,) containing quantized values
	    scales: float16 array of shape (nb,) containing scaling factors
	Returns:
	    float32 array of shape (k,) containing dequantized values
	"""
	scales = scales.astype(jnp.float32)
	dequantized = quants * scales
	return dequantized


@partial(jax.jit, static_argnames=["platform"])
def quantize_row_q8_0(array, platform):
	match platform:
		case "triton":
			return quantize_row_q8_triton_call(array)
		case "jax":
			return _mu_quantize_row_q8_0(array)
		case _:
			raise NotImplementedError(f"quantize_row_q8_0 not implemented for {platform}")


@partial(jax.jit, static_argnames=["platform"])
def dequantize_row_q8_0(quants, scales, platform):
	match platform:
		case "triton":
			return dequantize_row_q8_triton_call(quants, scales)
		case "jax":
			return _mu_dequantize_row_q8_0(quants, scales)
		case _:
			raise NotImplementedError(f"dequantize_row_q8_0 not implemented for {platform}")


@dataclass
class Array8Bit(core.ImplicitArray):
	"""
	Custom 8-bit Quantized Array for efficient manipulation of JAX arrays.

	This class provides methods for quantizing and dequantizing JAX arrays to 8-bit
	representation, which can significantly reduce memory usage and potentially
	improve computation speed for certain operations.

	Attributes:
	    array_quantized (core.ArrayValue): The quantized array data.
	    scale (core.ArrayValue): Scaling factors for dequantization.
	    shape (tuple): Shape of the quantized array.
	    dtype (jnp.dtype): Original dtype of the array before quantization.

	Example:
	    >>> import jax
	    >>> import jax.numpy as jnp

	    >>> x = jax.random.normal(jax.random.key(0), (512, 64), dtype=jnp.float32)
	    >>> xp = jax.random.normal(jax.random.key(1), (64, 256), dtype=jnp.float32)

	    >>> quantized_x = Array8Bit.quantize(x)
	    >>> quantized_xp = Array8Bit.quantize(xp)

	    >>> @jax.jit
	    >>> @core.implicit_compact
	    >>> def f(a, b):
	    ...   return jnp.dot(a, b)

	    >>> result = f(x, xp)
	    >>> q_result = f(quantized_x, quantized_xp)

	    >>> print(jnp.allclose(result, q_result, rtol=1e-2, atol=1e-2))
	    True
	"""

	array_quantized: core.ArrayValue
	scale: core.ArrayValue
	platform: Literal["jax", "triton", "pallas"] = core.aux_field()

	def materialize(self) -> Array:
		"""
		Materialize the quantized array back to its original representation.

		Returns:
		    Array: The dequantized array in its original dtype.
		"""
		return self.dequantize(
			array_quantized=self.array_quantized,
			scale=self.scale,
			float_dtype=self.dtype,
			shape=self.shape,
			platform=self.platform,
		)

	@classmethod
	def quantize(
		cls,
		array: Array,
		dtype: Optional[jnp.dtype] = None,
		q8: int = 128,
		platform: Optional[Literal["jax", "triton", "pallas"]] = None,
		*_,
		**__,
	) -> "Array8Bit":
		"""
		Quantize a JAX array to 8-bit representation.

		Args:
		    array (Array): The input array to quantize.
		    dtype (jnp.dtype, optional): The desired dtype for the output. If None, uses the input array's dtype.

		Returns:
		    Array8Bit: The quantized array.
		"""

		if platform is None:
			platform = "jax" if jax.default_backend() != "gpu" else "triton"
		org_shape = array.shape
		q8 = min(q8, array.size)
		if q8 % array.size != 0:
			q8 = array.shape[-1]
		array = array.reshape(-1, q8)
		quants, scales = quantize_row_q8_0(array=array, platform=platform)

		return cls(
			array_quantized=quants,
			scale=scales,
			shape=org_shape,
			dtype=dtype or array.dtype,
			platform=platform,
		)

	@staticmethod
	def dequantize(
		array_quantized: Array,
		scale: Array,
		float_dtype: jnp.dtype,
		shape: Sequence[int],
		platform: Literal["jax", "triton", "pallas"],
	) -> Array:
		"""
		Dequantize an 8-bit array back to its original representation.

		Args:
		    array_quantized (Array): The quantized array data.
		    scale (Array): The scaling factors used in quantization.
		    float_dtype (jnp.dtype): The desired output dtype.
				shape (Sequence[int]): org array shape
		Returns:
		    Array: The dequantized array.
		"""

		array = (
			dequantize_row_q8_0(
				quants=array_quantized,
				scales=scale,
				platform=platform,
			)
			.reshape(shape)
			.astype(float_dtype)
		)

		return array

	def __repr__(self) -> str:
		return f"Array8Bit(quants={self.array_quantized}, shape={self.shape}, dtype={self.dtype})"

	@property
	def nbytes(self) -> int:
		"""
		Calculate the total number of bytes used by the quantized representation.

		Returns:
		    int: The number of bytes used.
		"""
		return self.array_quantized.nbytes + self.scale.nbytes

	def memory_savings(self) -> float:
		"""
		Calculate the memory savings compared to the original array.

		Returns:
		    float: The percentage of memory saved.
		"""
		original_size = jnp.prod(jnp.array(self.shape)) * jnp.dtype(self.dtype).itemsize
		return (1 - self.nbytes / original_size) * 100


ArrayType = Union[Array, Array8Bit]


@core.primitive_handler("dot_general")
def handle_dot_general(
	primitive,
	lhs: ArrayType,
	rhs: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's dot_general operation.

	Materializes Array8Bit inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    lhs (ArrayType): Left-hand side array.
	    rhs (ArrayType): Right-hand side array.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.dot_general operation.
	"""
	if isinstance(lhs, Array8Bit):
		lhs = lhs.materialize()
	if isinstance(rhs, Array8Bit):
		rhs = rhs.materialize()
	return lax.dot_general(lhs, rhs, *args, **kwargs)


@core.primitive_handler("add")
def handle_add(
	primitive,
	x: ArrayType,
	y: ArrayType,
):
	"""
	Custom handler for JAX's add operation.

	Materializes Array8Bit inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): First array to add.
	    y (ArrayType): Second array to add.

	Returns:
	    The result of lax.add operation.
	"""
	if isinstance(x, Array8Bit):
		x = x.materialize()
	if isinstance(y, Array8Bit):
		y = y.materialize()
	return lax.add(x, y)


@core.primitive_handler("reduce")
def handle_reduce(
	primitive,
	operand: ArrayType,
	init_value: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's reduce operation.

	Materializes Array8Bit inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    operand (ArrayType): The array to be reduced.
	    init_value (ArrayType): The initial value for the reduction.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.reduce operation.
	"""
	if isinstance(operand, Array8Bit):
		operand = operand.materialize()
	if isinstance(init_value, Array8Bit):
		init_value = init_value.materialize()
	return lax.reduce(operand, init_value, *args, **kwargs)


@core.primitive_handler("mul")
def handle_mul(
	primitive,
	x: ArrayType,
	y: ArrayType,
):
	"""
	Custom handler for JAX's mul operation.

	Materializes Array8Bit inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): First array to multiply.
	    y (ArrayType): Second array to multiply.

	Returns:
	    The result of lax.mul operation.
	"""
	if isinstance(x, Array8Bit):
		x = x.materialize()
	if isinstance(y, Array8Bit):
		y = y.materialize()
	return lax.mul(x, y)


@core.primitive_handler("transpose")
def handle_transpose(
	primitive,
	operand: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's transpose operation.

	Materializes Array8Bit input before performing the operation.
	Re-quantizes the result if the input was Array8Bit.

	Args:
	    primitive: The JAX primitive being handled.
	    operand (ArrayType): The array to be transposed.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.transpose operation, potentially re-quantized.
	"""
	original_quantized = False
	if isinstance(operand, Array8Bit):
		operand = operand.materialize()
		original_quantized = True
	operand = lax.transpose(operand, *args, **kwargs)
	if original_quantized:
		operand = Array8Bit.quantize(operand, dtype=operand.dtype)
	return operand


@core.primitive_handler("conv_general_dilated")
def handle_conv(
	primitive,
	lhs: ArrayType,
	rhs: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's conv_general_dilated operation.

	Materializes Array8Bit inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    lhs (ArrayType): Left-hand side array (input).
	    rhs (ArrayType): Right-hand side array (kernel).
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.conv operation.
	"""
	if isinstance(lhs, Array8Bit):
		lhs = lhs.materialize()
	if isinstance(rhs, Array8Bit):
		rhs = rhs.materialize()
	return lax.conv_general_dilated(lhs, rhs, *args, **kwargs)


@core.primitive_handler("max")
def handle_max(
	primitive,
	x: ArrayType,
	y: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's max operation.

	Materializes Array8Bit inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): First array for max comparison.
	    y (ArrayType): Second array for max comparison.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.max operation.
	"""
	if isinstance(x, Array8Bit):
		x = x.materialize()
	if isinstance(y, Array8Bit):
		y = y.materialize()
	return lax.max(x, y, *args, **kwargs)


@core.primitive_handler("exp")
def handle_exp(
	primitive,
	x: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's exp operation.

	Materializes Array8Bit input before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): The array to apply exponential to.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.exp operation.
	"""
	if isinstance(x, Array8Bit):
		x = x.materialize()
	return lax.exp(x, *args, **kwargs)


@core.primitive_handler("log")
def handle_log(
	primitive,
	x: ArrayType,
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's log operation.

	Materializes Array8Bit input before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    x (ArrayType): The array to apply logarithm to.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.log operation.
	"""
	if isinstance(x, Array8Bit):
		x = x.materialize()
	return lax.log(x, *args, **kwargs)


@core.primitive_handler("reshape")
def handle_reshape(
	primitive: Primitive,
	operand: ArrayType,
	**kwargs: Any,
):
	"""
	Custom handler for JAX's reshape operation.

	This function handles reshaping for both regular arrays and Array8Bit quantized arrays.
	It materializes ArrayNF4 input before reshaping and re-quantizes the result if the input was ArrayNF4.

	Args:
	    primitive (Primitive): The JAX primitive being handled.
	    operand (ArrayType): The array to be reshaped.
	    new_sizes (Tuple[int, ...]): The desired new shape of the array.
	    dimensions (Tuple[int, ...], optional): The order in which dimensions should be permuted before reshaping.
	    **kwargs: Additional keyword arguments for the reshape operation.

	Returns:
	    ArrayType: The reshaped array, potentially re-quantized if the input was Array8Bit.

	Raises:
	    ValueError: If the new shape is not compatible with the original array's size.
	"""
	original_quantized = isinstance(operand, Array8Bit)

	if original_quantized:
		operand = operand.materialize()

	try:
		reshaped = lax.reshape(operand, **kwargs)
	except ValueError as e:
		raise ValueError(
			f"Reshape operation failed: {str(e)}. "
			f"Ensure the new shape {kwargs} is compatible with the original array size."
		) from e
	if original_quantized:
		reshaped = Array8Bit.quantize(reshaped, dtype=reshaped.dtype)
	return reshaped


@core.primitive_handler("concatenate")
def handle_concatenate(
	primitive,
	operands: Sequence[ArrayType],
	*args,
	**kwargs,
):
	"""
	Custom handler for JAX's concatenate operation.

	Materializes Array8Bit inputs before performing the operation.

	Args:
	    primitive: The JAX primitive being handled.
	    operands (Sequence[ArrayType]): Sequence of arrays to concatenate.
	    *args: Variable length argument list.
	    **kwargs: Arbitrary keyword arguments.

	Returns:
	    The result of lax.concatenate operation.
	"""
	materialized_operands = [
		op.materialize() if isinstance(op, Array8Bit) else op for op in operands
	]
	return lax.concatenate(materialized_operands, *args, **kwargs)


@core.primitive_handler("convert_element_type")
def convert_element_type(
	primitive,
	arg: ArrayType,
	**params,
) -> ArrayType:
	"""Handle element type conversion for Array8Bit."""
	result = jax.tree_util.tree_map(
		partial(core.default_handler, primitive, **params), arg
	)
	result.dtype = params["new_dtype"]
	return result


@core.primitive_handler("broadcast_in_dim")
def handle_broadcast_in_dim(
	primitive,
	operand: ArrayType,
	*args,
	**kwargs,
) -> ArrayType:
	"""Handle broadcast_in_dim for Array8Bit."""
	original_quantized = isinstance(operand, Array8Bit)
	array = operand
	if original_quantized:
		array = operand.materialize()
	result = jax.lax.broadcast_in_dim(array, *args, **kwargs)
	if original_quantized:
		result = Array8Bit.quantize(result, dtype=operand.dtype)
	return result


@core.primitive_handler("gather")
def handle_gather(
	primitive,
	operand: ArrayType,
	*args,
	**kwargs,
) -> ArrayType:
	"""Handle gather for Array8Bit."""
	original_quantized = isinstance(operand, Array8Bit)
	array = operand
	if original_quantized:
		array = operand.materialize()
	result = jax.lax.gather(array, *args, **kwargs)
	return result
